import functools
from pathlib import Path
from typing import Any

import yaml

from clippy.config import ClippyConfig
from clippy.kubernetes import KubernetesCluster
from clippy.puppet import PuppetHosts
from clippy.remote import Remote, RemoteHosts


class Clippy:
    """Provides an interface for the cookbooks to the Clippy API."""

    def __init__(self, config: ClippyConfig) -> None:
        self._config = config

    @functools.lru_cache(maxsize=None)
    def config(self, name: str) -> dict[str, Any]:
        with (self._config.config_dir / f"{name}.yaml").open("r") as f:
            config = yaml.safe_load(f)
        return config

    def kubernetes_cluster(self, name: str) -> KubernetesCluster:
        return KubernetesCluster(
            name=name,
            remote=self.remote,
            kubectl_host=self.remote.query("P{P:kubernetes::management}", single=True),
        )

    def puppet_hosts(self, fqdn: str) -> PuppetHosts:
        puppetserver_primary = self.config("puppet").get("puppetserver_primary")
        return PuppetHosts(
            fqdn=fqdn,
            remote=self.remote.query(f"D{{{fqdn}}}", single=True),
            installer_remote=self.remote_instance("installer").query(f"D{{{fqdn}}}", single=True),
            puppet_ca_remote=self.remote.query(f"D{{{puppetserver_primary}}}", single=True),
        )

    @functools.lru_cache(maxsize=None)
    def remote_instance(self, name: str = "default") -> Remote:
        return Remote(
            config_path=Path(self.config("remote").get("cumin_configs").get(name)),
        )

    @property
    def remote(self) -> Remote:
        return self.remote_instance()

    def remote_hosts(self, selector: str) -> RemoteHosts:
        return self.remote.query(selector=selector)
