import contextlib
import json
from argparse import ArgumentTypeError
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

from clippy.remote import Remote, RemoteHosts, RemoteResult


@dataclass
class KubernetesNodeInfo:
    name: str
    kubelet_version: str


class KubernetesWorker:
    def __init__(self, name: str, cluster: "KubernetesCluster") -> None:
        self._name = name
        self._cluster = cluster

    def depool(self, grace_period: int = 60) -> None:
        self._cluster.run_kubectl(
            command=f"drain {self._name} --ignore-daemonsets --delete-emptydir-data --grace-period={grace_period}",
            show_output=True,
        )

    def repool(self) -> None:
        self._cluster.run_kubectl(
            command=f"uncordon {self._name}",
            show_output=True,
        )

    def describe(self) -> KubernetesNodeInfo:
        object = self._cluster.get_object("node", self._name)
        return KubernetesNodeInfo(
            name=object["spec"]["name"],
            kubelet_version=object["status"]["nodeInfo"]["kubeletVersion"],
        )

    @contextlib.contextmanager
    def depooled(self, grace_period: int = 60) -> Generator[None, None, None]:
        self.depool(grace_period=grace_period)
        yield
        self.repool()


class KubernetesCluster:
    def __init__(self, name: str, remote: Remote, kubectl_host: RemoteHosts) -> None:
        self.name = name
        self._remote = remote

        self._kubectl_host = kubectl_host
        self._kubectl_base = f"KUBECONFIG=/etc/kubernetes/config.yaml kubectl --context={self.name}"

    def worker(self, fqdn: str) -> KubernetesWorker:
        name = fqdn.split(".")[0]
        return KubernetesWorker(
            name=name,
            cluster=self,
        )

    def run_kubectl(
        self,
        command: str,
        *,
        require_success: bool = True,
        show_output: bool = False,
        show_progress: bool = False,
    ) -> RemoteResult:
        return self._kubectl_host.execute(
            command=f"{self._kubectl_base} {command}",
            require_success=require_success,
            show_output=show_output,
            show_progress=show_progress,
        )

    def get_object(self, obj_type: str, name: str) -> dict[str, Any]:
        result = self.run_kubectl(f"get {obj_type} {name} -o json").get_output()
        return json.loads(result)


def validate_version(version: str) -> str:
    """Argparse type validator for Kubernetes versions."""
    parts = [part.strip() for part in version.split(".") if part.strip() != "" and part.strip().isnumeric()]
    if len(parts) != 3:
        raise ArgumentTypeError(f"Expected version in minor.major.patch format, got '{version}'")
    return ".".join(parts)
