#!/usr/bin/env python3

import argparse
import importlib
import logging
import sys
from pathlib import Path
from typing import Type

from clippy.api import Clippy
from clippy.config import ClippyConfig, load_config
from clippy.cookbook import Cookbook
from clippy.interactive import ClippyLogFormatter

RETCODE_COOKBOOK_NOT_FOUND = 90


logger = logging.getLogger(__name__)


def discover_cookbooks(config: ClippyConfig) -> dict[str, Type[Cookbook]]:
    cookbooks = {}

    for namespace, package in config.cookbook_namespaces.items():
        cookbook_directory = Path(importlib.import_module(package).__file__).parent
        for file in cookbook_directory.glob("**/*.py"):
            if file.name.startswith("_"):
                continue
            file_subpath = str(file.relative_to(cookbook_directory)).replace("/", ".").replace(".py", "")
            module_name = f"{package}.{file_subpath}"
            cookbook_name = f"{namespace}.{file_subpath.replace('_', '-')}"

            module = importlib.import_module(module_name)

            attrs = [getattr(module, attr) for attr in dir(module)]
            classes = [
                attr
                for attr in attrs
                if isinstance(attr, type) and issubclass(attr, Cookbook) and attr.__module__ == module.__name__
            ]

            if len(classes) == 1:
                cookbooks[cookbook_name] = classes[0]

    return cookbooks


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="clippy: automation toolkit", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable verbose output",
    )

    parser.add_argument("--config", type=Path, default=Path("/etc/clippy/config.yaml"), help="use specific config file")
    parser.add_argument("-l", "--list", action="store_true", help="lists available cookbooks")
    parser.add_argument("--list-noformat", action="store_true", help=argparse.SUPPRESS)

    parser.add_argument("cookbook_name", metavar="COOKBOOK", nargs="?", help="Cookbook to execute", default="")
    parser.add_argument(
        "cookbook_args",
        metavar="ARGS",
        nargs=argparse.REMAINDER,
        help="Cookbook arguments",
    )

    return parser.parse_args()


def setup_logging(verbose: bool):
    logging.captureWarnings(True)
    root = logging.getLogger()
    root.setLevel(logging.WARN)

    stream = logging.StreamHandler()
    stream.setFormatter(ClippyLogFormatter())
    root.addHandler(stream)

    clippy = logging.getLogger("clippy")
    clippy.setLevel(logging.DEBUG if verbose else logging.INFO)


def main() -> None:
    """Main CLI entry point"""
    args = parse_args()
    config = load_config(args.config)
    cookbooks = discover_cookbooks(config)

    setup_logging(args.verbose)

    if args.list or len(args.cookbook_name) == 0:
        filtered_cookbooks = (
            cookbooks
            if len(args.cookbook_name) == 0
            else {name: cookbook for name, cookbook in cookbooks.items() if name.startswith(args.cookbook_name)}
        )

        filtered_cookbooks = {name: cookbook for name, cookbook in filtered_cookbooks.items() if not cookbook.unlisted}

        if not args.list_noformat:
            if len(filtered_cookbooks) == 0:
                logger.warning("There are no cookbooks matching the filter.")
                sys.exit(0)

            logger.info("Available cookbooks:")

        for name, cookbook in sorted(filtered_cookbooks.items()):
            if args.list_noformat:
                print(name)
            else:
                logger.info(f" {name} ({cookbook.title})")

        sys.exit(0)

    if args.cookbook_name not in cookbooks:
        logger.error("No such cookbook %s", args.cookbook_name)
        sys.exit(RETCODE_COOKBOOK_NOT_FOUND)

    clippy = Clippy(config)
    cookbook = cookbooks[args.cookbook_name](args.cookbook_name, clippy)
    cookbook.args = cookbook.argument_parser().parse_args(args.cookbook_args)

    retval = cookbook.execute()

    if retval is not None:
        sys.exit(retval)

    sys.exit(0)
