import json
import logging
from typing import Any

from clippy.remote import RemoteCommandError, RemoteHosts

logger = logging.getLogger(__name__)

CERTIFICATE_FINGERPRINT_MAGIC = r"sha256sum | awk '{ print $1 }' | sed 's/\(..\)/\1:/g; s/:$//; s/./\U&/g;'"


class PuppetHosts:
    def __init__(
        self, fqdn: str, remote: RemoteHosts, installer_remote: RemoteHosts, puppet_ca_remote: RemoteHosts
    ) -> None:
        self._fqdn = fqdn
        self._remote = remote
        self._installer_remote = installer_remote
        self._puppet_ca_remote = puppet_ca_remote

    def clean(self) -> None:
        self._puppet_ca_remote.execute(command=f"puppet node deactivate {self._fqdn}", show_output=True)
        self._puppet_ca_remote.execute(command=f"puppet node clean {self._fqdn}", show_output=True)

    def _submit_csr(self) -> str:
        self._installer_remote.execute(command="puppet ssl submit_request --waitforcert 0")
        for _, output in self._installer_remote.execute(
            command=f"openssl req -in /var/lib/puppet/ssl/certificate_requests/{self._fqdn}.pem -outform der | {CERTIFICATE_FINGERPRINT_MAGIC}",
            show_output=True,
        ).results:
            for line in output.message().decode().splitlines():
                return line
        raise Exception("Failed to parse fingerprint")

    def _get_cert_info(self) -> dict[str, Any]:
        output = self._puppet_ca_remote.execute(
            command=f"puppetserver ca list --format json --certname {self._fqdn}"
        ).get_output()

        response = json.loads(output)
        key = list(response.keys())[0]
        if key == "missing":
            raise Exception("Did not find certificate info")
        metadata = response[key][0]
        if metadata["name"] != self._fqdn:
            raise Exception("Got metadata for wrong server?")
        return metadata

    def _sign(self, fingerprint: str):
        cert_info = self._get_cert_info()
        if (
            cert_info["state"] != "requested"
            or cert_info["dns_alt_names"] != []
            or cert_info["subject_alt_names"] != []
            or cert_info["authorization_extensions"] != {}
            or cert_info["fingerprint"] != fingerprint
        ):
            raise Exception("Found invalid CSR")
        self._puppet_ca_remote.execute(command=f"puppetserver ca sign --certname {self._fqdn}")

    def sign_certificates(self):
        # TODO: do not hardcode this
        self._installer_remote.execute(command="puppet config set server puppet.discovery.majava.org")

        try:
            csr_fingerprint = self._submit_csr()
        except RemoteCommandError:
            self._puppet_ca_remote.execute(
                command=f"puppetserver ca clean --certname {self._fqdn}", require_success=False
            )
            csr_fingerprint = self._submit_csr()
        self._sign(csr_fingerprint)
