from __future__ import annotations

import logging
import os
import time
from datetime import datetime
from pathlib import Path

from ClusterShell.MsgTree import MsgTreeElem
from ClusterShell.NodeSet import NodeSet
from cumin import Config
from cumin.query import Query
from cumin.transport import Transport
from cumin.transports import Target
from cumin.transports.clustershell import NullReporter, TqdmReporter

from clippy.exceptions import AutomationError

logger = logging.getLogger(__name__)


class RemoteError(AutomationError):
    pass


class RemoteQueryError(RemoteError):
    pass


class Remote:
    def __init__(self, config_path: Path) -> None:
        self.cumin_config = Config(str(config_path))

    def query(self, selector: str, single: bool = False) -> "RemoteHosts":
        hosts = Query(self.cumin_config).execute(selector)
        if single:
            if len(hosts) == 0:
                raise RemoteQueryError(f"Selector '{selector}' matches no hosts")
            elif len(hosts) > 1:
                host = hosts[0]
                hosts = NodeSet.fromlist([host])

        return RemoteHosts(cumin_config=self.cumin_config, hosts=hosts)


class RemoteResult:
    def __init__(self, results: list[tuple[NodeSet, MsgTreeElem]]) -> None:
        self.results = results

    def get_output(self) -> str:
        if len(self.results) != 1:
            raise RemoteError(f"RemoteResult.get_output() expects exactly 1 result, got {len(self.results)}")
        return self.results[0][1].message().decode()


class RemoteCommandError(RemoteError):
    def __init__(self, message: str, ret: int, results: RemoteResult) -> None:
        super().__init__(f"{message} (exit_code={ret})")
        self.exit_code = ret
        self.results = results


class RemoteRebootError(RemoteError):
    pass


class RemoteHosts:
    def __init__(self, *, cumin_config: Config, hosts):
        self.cumin_config = cumin_config
        self.hosts = hosts

    def execute(
        self,
        *,
        command: str,
        require_success: bool = True,
        show_output: bool = False,
        show_progress: bool = False,
    ) -> RemoteResult:
        target = Target(self.hosts)

        if os.environ.get("CLIPPY_SUDO") == "1":
            command = f"sudo -i {command}"

        worker = Transport.new(self.cumin_config, target)
        worker.commands = [command]
        worker.handler = "sync"
        worker.progress_bars = show_progress

        if show_output:
            worker.reporter = TqdmReporter
        else:
            worker.reporter = NullReporter

        logger.debug(
            "Executing command '%s' on %d hosts: %s",
            command,
            len(target.hosts),
            str(target.hosts),
        )

        ret = worker.execute()
        result = RemoteResult(results=[result for result in worker.get_results()])
        if ret != 0 and require_success:
            raise RemoteCommandError("Command failed", ret, results=result)

        return result

    def _wait_reboot_since(self, since: datetime):
        for i in range(10):
            try:
                uptimes = self.execute(command="cat /proc/uptime").results
            except RemoteCommandError:
                time.sleep(5)
                continue
            for nodeset, output in uptimes:
                delta = (datetime.utcnow() - since).total_seconds()
                uptime = float(output.message().decode().strip().split()[0])
                if uptime >= delta:
                    logger.info(f"Hosts %s are up for %s seconds (expected less than %s)", nodeset, uptime, delta)
                    time.sleep(5)
                    continue
            return
        raise RemoteRebootError("Gave up waiting for a reboot")

    def reboot(self) -> None:
        now = datetime.utcnow()
        self.execute(command="reboot-host")
        self._wait_reboot_since(since=now)

    def __len__(self) -> int:
        return len(self.hosts)

    def __str__(self):
        return str(self.hosts)
